@@ -43,6 +43,7 @@ func TestSimRandomResolve(t *testing.T) {
 
 	// A new node joins every 10s.
 	launcher := time.NewTicker(10 * time.Second)
+	defer launcher.Stop()
 	go func() {
 		for range launcher.C {
 			net := sim.launchNode(false)
@@ -50,12 +51,11 @@ func TestSimRandomResolve(t *testing.T) {
 			if err := net.SetFallbackNodes([]*Node{bootnode.Self()}); err != nil {
 				panic(err)
 			}
-			fmt.Printf("launched @ %v: %x\n", time.Now(), net.Self().ID[:16])
+			t.Logf("launched @ %v: %x\n", time.Now(), net.Self().ID[:16])
 		}
 	}()
 
 	time.Sleep(3 * time.Hour)
-	launcher.Stop()
 	sim.shutdown()
 	sim.printStats()
 }
@@ -196,6 +196,7 @@ func randomResolves(t *testing.T, s *simulation, net *Network) {
 	}
 
 	timer := time.NewTimer(randtime())
+	defer timer.Stop()
 	for {
 		select {
 		case <-timer.C:
@@ -294,15 +295,6 @@ func (s *simulation) launchNode(log bool) *Network {
 	return net
 }
 
-func (s *simulation) dropNode(id NodeID) {
-	s.mu.Lock()
-	n := s.nodes[id]
-	delete(s.nodes, id)
-	s.mu.Unlock()
-
-	n.Close()
-}
-
 type simTransport struct {
 	joinTime   time.Time
 	sender     NodeID
@@ -358,22 +350,6 @@ func (st *simTransport) sendPing(remote *Node, remoteAddr *net.UDPAddr, topics [
 	return hash
 }
 
-func (st *simTransport) sendPong(remote *Node, pingHash []byte) {
-	raddr := remote.addr()
-
-	st.sendPacket(remote.ID, ingressPacket{
-		remoteID:   st.sender,
-		remoteAddr: st.senderAddr,
-		hash:       st.nextHash(),
-		ev:         pongPacket,
-		data: &pong{
-			To:         rpcEndpoint{IP: raddr.IP, UDP: uint16(raddr.Port), TCP: 30303},
-			ReplyTok:   pingHash,
-			Expiration: uint64(time.Now().Unix() + int64(expiration)),
-		},
-	})
-}
-
 func (st *simTransport) sendFindnodeHash(remote *Node, target common.Hash) {
 	st.sendPacket(remote.ID, ingressPacket{
 		remoteID:   st.sender,
